# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import abc
from urllib.parse import urlparse, urlencode, parse_qs, quote
from requests import Request
from arjuna.tpi.error import HttpRequestCreationError
from arjuna.tpi.parser.json import Json, JsonDict, JsonList
from arjuna.tpi.parser.html import Html
from arjuna.tpi.engine.asserter import AsserterMixIn
import time

from .message import HttpMessage
from .response import HttpResponse

class HttpRequest(HttpMessage):
    '''
        Encapsulates HTTP response message.

        Arguments:
            session: `HttpSession` object which created this `HttpRequest`.
            request: `requests` library's Request object wrapped by this class.

        Keyword Arguments:
            label: Label for this request. If available, it is used in Reports and logs.
            xcodes: Expected Status Code(s)
            strict: If True in case of unexpected status code, an AssertionError is raised, else HttpUnexpectedStatusCodeError is raised.
            allow_redirects: If True, redirections are allowed for the HTTP message. Default is True.
            timeout: How long to wait for the server to send data before giving up.
            cookies: Cookie dictionary
    '''

    def __init__(self, session, request, label=None, xcodes=None, strict=False, allow_redirects=True, timeout=None, cookies=None):
        super().__init__(request)
        self.__session = session
        self.__request = request
        req_repr = "{} {}".format(self.method, self.url)
        self.__label = label and label or req_repr
        printable_label = label and self.__label + "::" + req_repr or req_repr
        printable_label = len(printable_label) > 119 and printable_label[:125] + "<SNIP>" or printable_label
        self.__label = printable_label
        self.__strict = strict
        self.__allow_redirects = allow_redirects
        self.__timeout = timeout
        self.__cookies = cookies

    @property
    def _request(self):
        return self.__request

    @property
    def allow_redirects(self):
        '''
        True if redirects are allowed for this request.
        '''
        return self.__allow_redirects

    @property
    def timeout(self):
        '''
        Timeout for this request.
        '''
        return self.__timeout

    @property
    def label(self) -> str:
        '''
            Label for this request object.
        '''
        return self.__label

    @property
    def query_params(self) -> dict:
        '''
            URL Query Parameters for this request object.
        '''
        return parse_qs(self.url)

    @property
    def url(self) -> str:
        '''
            URL correspnding to this request message.
        '''
        return self.__request.url

    @property
    def method(self) -> str:
        '''
            HTTP Method/Verb used by this request.
        '''
        return self.__request.method

    @property
    def xcodes(self) -> set:
        '''
        Expected status codes for this request.
        '''
        return self.__xcodes

    @property
    def strict(self) -> bool:
        '''
        True if sending this request should raise an Assertion error in case of unexpected status code.
        '''
        return self.__strict

    @property
    def text(self):
        '''
            Content of this request message as plain text.
        '''
        return self.__request.body

    content = text

    __OUT ='''{}
{}{}'''

    @classmethod
    def repr_as_str(cls, *, method, url, headers, content=None):
        if content:
            content = '\n\n{}\n'.format(content)
        else:
            content = ""
        return cls.__OUT.format(
            method + ' ' + url,
            '\n'.join('{}: {}'.format(k, v) for k, v in headers.items()),
            content
        ).strip()

    def __str__(self):
        content = self.text
        if isinstance(content, bytes):
            content = content.decode()
        return self.repr_as_str(
            method = self.__request.method,
            url = self.url,
            headers = self.__request.headers,
            content = content
        )


class _HttpRequest(HttpRequest):

    def __init__(self, session, url, method, label=None, content=None, xcodes=None, strict=False, headers=None, cookies=None, allow_redirects=True, auth=None, timeout=None, pretty_url=False, query_params=None, **named_query_params):
        self.__session = session
        self.__method = method.upper()
        self.__url = url
        self.__content = None
        if content is not None:
            if type(content) in {str, dict, list, tuple} or isinstance(content, JsonDict) or isinstance(content, JsonList):
                self.__content = self.__session.request_content_handler(content)
            else:
                self.__content = content
        self.__xcodes = None
        if xcodes is not None:
            self.__xcodes = type(xcodes) in {set, list, tuple} and xcodes or {xcodes}
        self.__query_params = query_params
        if self.__query_params is None:
            self.__query_params = dict()
        self.__query_params.update(named_query_params)
        self.__pretty_url = pretty_url
        self.__headers = {}
        self.__headers.update(session.headers)
        if headers:
            self.__headers.update(headers)
        self.__cookies = dict()
        self.__cookies.update(self.__session.cookies)
        if cookies is not None:
            self.__cookies.update(cookies)
        self.__auth = auth is not None and auth or self.__session.auth
        self.__prepare_headers()
        self.__req = self.__build_request()
        super().__init__(self.__session, self.__req, label=label, xcodes=self.__xcodes, strict=strict, allow_redirects=allow_redirects, timeout=timeout, cookies=self.__cookies)

    def __prepare_headers(self):
        if self.__method in {'POST', 'PUT', 'PATCH', 'OPTIONS'}:
            self.__headers['Content-Type'] = self.__content.type
        if self.__method in {'GET', 'HEAD', 'DELETE'}:
            if 'Content-Type' in self.__headers:
                del self.__headers['Content-Type']

    def __build_request(self):
        parsed_uri = urlparse(self.__url)
        #self.__headers['Host'] = parsed_uri.netloc
        if self.__pretty_url:
            query = "/".join([quote(f"{k}/{v}") for k,v in self.__query_params.items()])
            url = self.__url + query
            query_params = None
        else:
            url = self.__url
            query_params = self.__query_params

        try:
            if self.__content:
                data = self.__content.content
            else:
                data = None
            req = Request(self.__method, url, data=data, headers=self.__headers, params=query_params, cookies=self.__cookies, auth=self.__auth)
            return req.prepare()
        except Exception as e:
            raise HttpRequestCreationError("{} : {}".format(e.__class__.__name__, str(e)))




# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
import os
import re
import sys
from urllib3.util import parse_url
from arjuna.tpi.constant import *
from arjuna.core.constant import *
from arjuna.core.types.constants import *

class Validator:
    VNREGEX = r'^([a-zA-Z][a-zA-Z_0-9]{2,50})$'
    VNREGEX_TEXT = '''
    Name must be a string of length 3-50 starting with a letter, followed by letters, digits or _ (underscore).
    '''

    @classmethod
    def raise_exc(cls, input, msg=None):
        curframe = inspect.currentframe()
        callframe = inspect.getouterframes(curframe, 2)
        caller = callframe[1][3]
        msg = msg is not None and msg + "." or ""
        raise Exception("{} is not {}. {}".format(input, caller, msg))

    @classmethod
    def str(cls, input):
        if type(input) not in {str, int, float, bool}:
            cls.raise_exc(input)
        return str(input)

    @classmethod
    def bool(cls, input):
        if type(input) is not bool:
            if type(input) is str:
                if input.lower() in TRUES:
                    return True
                elif input.lower() in FALSES:
                    return False
                else:
                    cls.raise_exc(input)
            else:
                cls.raise_exc(input)
        return input

    @classmethod
    def int(cls, input):
        if type(input) is not int:
            if type(input) is str:
                try:
                    return int(input)
                except:
                    cls.raise_exc(input)
            else:
                cls.raise_exc(input)
        return input

    @classmethod
    def float(cls, input):
        if type(input) is not float:
            if type(input) is int:
                return float(input)
            elif type(input) is str:
                try:
                    return float(input)
                except:
                    cls.raise_exc(input)
            else:
                cls.raise_exc(input)
        return input

    @classmethod
    def absolute_dir_path_present(cls, input):
        if type(input) is not str:
            cls.raise_exc(input)
        elif not os.path.isabs(input) or not os.path.isdir(input):
            cls.raise_exc(input)
        return input

    @classmethod
    def absolute_dir_path(cls, input):
        if type(input) is not str:
            cls.raise_exc(input)
        elif os.path.exists(input):
            if not os.path.isabs(input) or not os.path.isdir(input):
                cls.raise_exc(input)
        # else:
        #     os.makedirs(input)
        return input

    @classmethod
    def absolute_file_path(cls, input):
        if type(input) is not str:
            cls.raise_exc(input)
        elif not os.path.isabs(input) or not os.path.isfile(input):
            cls.raise_exc(input)
        return input

    @classmethod
    def guiauto_automator_name(cls, input):
        try:
            return GuiAutomatorName[input.upper()]
        except:
            cls.raise_exc(input)

    @classmethod
    def guiauto_context_name(cls, input):
        try:
            return GuiAutomationContext[input.upper()]
        except Exception as e:
            print(e)
            cls.raise_exc(input)

    @classmethod
    def logging_level(cls, input):
        try:
            return LoggingLevel[input.upper()]
        except Exception as e:
            print(e)
            cls.raise_exc(input)

    @classmethod
    def locale(cls, input):
        try:
            return Locale[input.upper()]
        except Exception as e:
            print(e)
            cls.raise_exc(input)

    @classmethod
    def str_list(cls, input):
        if type(input) is not list:
            cls.raise_exc(input)
        else:
            s = {type(i) for i in input}
            if s != {type("")}:
                cls.raise_exc(input)
        return input

    @classmethod
    def allowed_log_contexts(cls, input):
        if type(input) is str:
            input = [i.strip().lower() for i in input.split(",")]
            input = set(input)
        elif type(input) in {set, list, tuple}:
            input = set([i.lower() for i in input])
        else:
            cls.raise_exc(input)    

        input.add("default")
        return input 

    @classmethod
    def str_or_strlist(cls, input):
        if type(input) is str:
            input = [i.strip().lower() for  i in input.split(",")]
        elif type(input) in {set, list, tuple}:
            input = [i.lower() for i in input if i not in input]
        else:
            cls.raise_exc(input)
        return list(dict.fromkeys(input)) 

    @classmethod
    def actor_mode(cls, input):
        try:
            return SetuActorMode[input.upper()]
        except:
            cls.raise_exc(input)

    @classmethod
    def browser_name(cls, input):
        try:
            if isinstance(input, BrowserName):
                return input
            return BrowserName[input.upper()]
        except:
            cls.raise_exc(input)

    @classmethod
    def desktop_os(cls, input):
        try:
            return DesktopOS[input.upper()]
        except:
            cls.raise_exc(input)

    @classmethod
    def report_formats(cls, input):
        l = None
        try:
            if type(input) is str:
                return cls.report_formats([i.upper() for i in input.split(",")])
            elif type(input) in {list, tuple}:
                return [ReportFormat[i.upper()] for i in input]
        except:
            cls.raise_exc(input)

    @classmethod
    def mobile_os(cls, input):
        try:
            return MobileOsName[input.upper()]
        except:
            cls.raise_exc(input)

    @classmethod
    def positive_int(cls, input):
        if type(input) is int and input <= 1:
            cls.raise_exc(input)
        if type(input) is not int:
            i = cls.int(input)
            if i <= 1:
                cls.raise_exc(input)
            else:
                return i
        return input

    @classmethod
    def positive_float(cls, input):
        if type(input) is float and input < 0.1:
            cls.raise_exc(input)

        if type(input) is not float:
            i = cls.float(input)
            if i <= 1:
                cls.raise_exc(input)
            else:
                return i
        return input

    @classmethod
    def web_url(cls, input):
        def check_scheme():
            try:
                return parse_url(input).scheme in {'http', 'https'}
            except Exception as e:
                return False
        if  type(input) is not str or not check_scheme():
            cls.raise_exc(input)
        return input

    @classmethod
    def name(cls, input):
        if not re.match(cls.VNREGEX, input):
            print('Invalid name provided.', file=sys.stderr)
            print(cls.VNREGEX_TEXT, file=sys.stderr)
            cls.raise_exc(input, msg=cls.VNREGEX_TEXT)
        return input


'''
runid(self, prop_path, config_value, purpose, visible):
        m = re.match(r"^[a-zA-Z0-9\-_]{3,30}$", config_value)
'''


from enum import Enum, auto

class WithType(Enum):
    ID = auto()
    NAME = auto()
    CLASSES = auto()
    LINK = auto()
    FLINK = auto()
    TAG = auto()
    XPATH = auto()
    SELECTOR = auto()

    TEXT = auto()
    FTEXT = auto()
    BTEXT = auto()
    ETEXT = auto()
    TITLE = auto()
    ATTR = auto()
    FATTR = auto()
    BATTR = auto()
    EATTR = auto()
    ATTR_WORD = auto()
    VALUE = auto()
    POINT = auto()
    JS = auto()

    INDEX = auto()
    WINDOW_TITLE = auto()
    WINDOW_PTITLE = auto()
    ELEMENT = auto()
    LABEL = auto()


class GenericLocateWith(Enum):
    ID = auto()
    NAME = auto() 
    XPATH = auto() 
    SELECTOR = auto()
    CLASSES = auto() 
    LINK = auto() 
    FLINK = auto()
    PLINK = auto() 
    TAG = auto()

    INDEX = auto()
    WINDOW_TITLE = auto()
    WINDOW_PTITLE = auto()
    ELEMENT = auto()
    POINT = auto()
    JS = auto() 

    # Translated to XPath
    TEXT = auto() 
    FTEXT = auto() 
    TITLE = auto() 
    VALUE = auto() 
    ATTR = auto()
    FATTR = auto()
    IMAGE_SRC = auto()
    IMAGE = auto()

    # Translated 1-1 to Selenium
    PARTIAL_LINK_TEXT = auto()
    LINK_TEXT = auto()
    TAG_NAME = auto()
    CSS_SELECTOR = auto()


# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
Arjuna Test Resources

The test resources markup provided by Arjuna includes easy-to-use decorators which wrap **pytest.fixture** decorator.

Note:
    For using any of the decorators in this module, the resource function must have the signature as **f(request)** with **request** as the first argument.
'''

import pytest
import itertools
import functools

from .test import My
from typing import Callable


def _simple_dec(func):
    @functools.wraps(func)
    def call_func(request, *args, **kwargs):
        from arjuna import log_info
        request_wrapper = My()
        request_wrapper.set_req_obj(request)
        request_wrapper.data = hasattr(request, "param") and request.param or None
        qual_name = request_wrapper.info.get_qual_name_with_data()
        if func.__name__ != 'group':
            log_info("(Setup) Begin fixture function: {}".format(qual_name))   
        yield from func(request_wrapper, *args, **kwargs)
        if func.__name__ != 'group':
            log_info("(Teardown) End fixture function: {}".format(qual_name))
        
    return call_func

def _repr_record(record):
    return str(record)

def __pytestfix(scope, drive_with=None, default=False):
    records = None
    ids = None
    if drive_with:
        records = drive_with.build("Fixture").all_records
        ids = [_repr_record(record) for record in records]

    def wrapper(func):
        orig_func = func
        if drive_with:
            return pytest.fixture(scope=scope, params=records, ids=ids, autouse=default)(_simple_dec(func))
        else:
            return pytest.fixture(scope=scope, autouse=default)(_simple_dec(func))
    return wrapper


def for_group(func: Callable=None, *, drive_with: 'DataSource'=None, default: bool=False) -> Callable:
    '''
        Decorator for session level test resource.

        Wraps **pytest.fixture** to create a fixture with scope=session and provides an Arjuna's decorated version of the function that is marked with **for_group** decorator.

        Args:
            func: A Function with signature **f(request)**. The name request is mandatory and enforced.

        Keyword Arguments:
            drive_with: (Optional) Used for data driven testing. Argument can be Arjuna Data Source.

                .. code-block:: python
                
                    @for_group(drive_with=<DS>)
                    def fix(request, data):
                        pass

            default: Should it be auto-applied?
    '''
    from arjuna import Arjuna

    if func is not None:
        return pytest.fixture(scope="session", autouse=default)(_simple_dec(func))
    else:
        return __pytestfix("session", drive_with=drive_with, default=default)


def for_module(func: Callable=None, *, drive_with: 'DataSource'=None, default: bool=False) -> Callable:
    '''
        Decorator for module level test fixture/resource.

        Wraps **pytest.fixture** to create a fixture with scope=module and provides an Arjuna's decorated version of the function that is marked with **for_module** decorator.

        Args:
            func: A Function with signature **f(request)**. The name request is mandatory and enforced.

        Keyword Arguments:
            drive_with: (Optional) Used for data driven testing. Argument can be Arjuna Data Source.

                .. code-block:: python
                
                    @for_module(drive_with=<DS>)
                    def fix(request, data):
                        pass

            default: Should it be auto-applied?
    '''
    from arjuna import Arjuna

    if func is not None:
        return pytest.fixture(scope="module", autouse=default)(_simple_dec(func))
    else:
        return __pytestfix("module", drive_with=drive_with, default=default)


def for_test(func: Callable=None, *, drive_with: 'DataSource'=None, default: bool=False) -> Callable:
    '''
        Decorator for test function level test fixture/resource.

        Wraps **pytest.fixture** to create a fixture with scope=function and provides an Arjuna's decorated version of the function that is marked with **for_test** decorator.

        Args:
            func: A Function with signature **f(request)**. The name request is mandatory and enforced.

        Keyword Arguments:
            drive_with: (Optional) Used for data driven testing. Argument can be Arjuna Data Source.

                .. code-block:: python
                
                    @for_test(drive_with=<DS>)
                    def fix(request):
                        pass

            default: Should it be auto-applied?
    '''
    from arjuna import Arjuna

    if func is not None:
        return pytest.fixture(scope="function", autouse=default)(_simple_dec(func))
    else:
        return __pytestfix("function", drive_with=drive_with, default=default)


@for_group(default=True)
def group(request):
    getattr(request.raw_request, "session").group_info = request.raw_request.param
    yield

@for_test(default=True)
def test_resources(request):
    from arjuna import Arjuna
    thread_container = Arjuna.get_report_metadata()
    set_node_id = thread_container.current_test_node_id  
    thread_container.current_test_node_id = request.raw_request.node.nodeid
    getattr(request.raw_request, "function").report_metadata = thread_container
    yield
    try:
        network_recorder = request.space.network_recorder
        network_recorder.register()
    except:
        pass
    thread_container.clear()


# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .generator import _gen

def _init(self, **kwargs):
    wrong_args = False
    provided_set = kwargs.keys()
    mandatory_args_missing = self._MANDATORY - provided_set
    if mandatory_args_missing:
        wrong_args = True
    default_arg_diff = kwargs.keys() - self._REFDICT.keys()
    if default_arg_diff:
        wrong_args = True

    if wrong_args:
        mandatory_args_missing = mandatory_args_missing and f" Missing mandatory args: {mandatory_args_missing}. Must: {self._MANDATORY}" or ""
        allowed = set(self._REFDICT.keys())
        default_arg_diff = default_arg_diff and f" Unexpected args: {default_arg_diff}. Allowed: {allowed}." or ""
        raise Exception(f"Wrong arguments.{mandatory_args_missing}{default_arg_diff}")

    def update_gen_data(key, value):
        if hasattr(value, '__call__'):
            value = value()
        elif isinstance(value, _gen):
            value = value.generate()

        vars(self)[key] = value

    final_dict = dict()
    final_dict.update(self._REFDICT)
    final_dict.update(kwargs)    

    for k,v in final_dict.items():
        update_gen_data(k,v)

def _attr(self, name):
    if name in self._REFDICT:
        return self._REFDICT[name]
    else:
        raise AttributeError("{} object does not have an attribute with name: {}.".format(self.__class__.__name__, name))

def _as_dict(self, remove_none=False):
    d = dict(vars(self))
    if remove_none:
        return {k:v for k,v in d.items() if v is not None}
    else:
        return d

def _str(self):
    return "{klass}({attrs})".format(
        klass = self.__class__.__name__,
        attrs = ", ".join("{}={}".format(k,v) for k,v in self.as_dict().items())
    )

def data_entity(entity_name, *attrs, **attrs_with_defaults):
    '''
        Create a new Data Entity class with provided name and attributes.

        Arguments:
            entity_name: The class name for this new Data Entity type.
            *attrs: Arbitrary names for Python attributes to be associated with objects of this entity.

        Keyword Arguments:
            **attrs_with_defaults: Arbitrary attributes to be associated with objects of this entity, with the defaults that are provided.

        Note:
            The defaults that are provided can be any of the following:
                * Any Python Object
                * A Python callable
                * Arjuna `generator`
                * Arjuna `composite`
    '''
    for attr in attrs:
        if type(attr) is not str:
            raise Exception(f"Data Entity fields must be valid Python name strings. Got: {attr}")

    final_attrs = []
    # In any multiple names in single string are provided like a namedtuple, added as individual names.
    for attr in attrs:
        final_attrs.extend(attr.split())
    namespace = dict()
    namespace['_MANDATORY'] = set(final_attrs)
    namespace['_DEFAULT'] = set(attrs_with_defaults.keys())
    namespace['_REFDICT'] = dict()
    namespace['_REFDICT'].update({f: None for f in final_attrs})
    namespace['_REFDICT'].update(attrs_with_defaults)
    namespace['__init__'] = _init
    namespace['__getattr__'] = _attr
    namespace['as_dict'] = _as_dict
    namespace['__str__'] = _str
    return type(entity_name, tuple(), namespace)


# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from arjuna.tpi.guiauto.base.locatable import Locatable
from arjuna.tpi.guiauto.base.single_widget import SingleGuiWidget
from arjuna.tpi.guiauto.base.container import GuiWidgetContainer
from arjuna.tpi.engine.asserter import AsserterMixIn
from arjuna.tpi.guiauto.model.gns import GNS
from arjuna.interact.gui.auto.finder import GuiElementFinder, GuiElementEmdFinder
from arjuna.tpi.error import *
from arjuna.core.error import *
from arjuna.tpi.tracker import track
from arjuna.tpi.guiauto.meta.formatter import GuiWidgetLocatorFormatter

@track("debug")
class GuiElement(AsserterMixIn, GuiWidgetContainer, Locatable, SingleGuiWidget):
    '''
        Represents a single element in GUI of any kind.

        Not meant to be directly created. It is created using calls from **Gui** object or **GuiNamespace** object of **Gui**.

        Arguments:
            gui: Gui object containing this element.
            wmd: **GuiElementMetaData** object for this element.
    '''

    def __init__(self, gui, wmd):
        AsserterMixIn.__init__(self)
        GuiWidgetContainer.__init__(self, gui._automator.config)
        Locatable.__init__(self, gui, wmd) #, parent, obj_name="GuiElement")
        SingleGuiWidget.__init__(self, gui, wmd)
        self.__gns = GNS(self, gui._gui_def)
        self.__finder = GuiElementFinder(self)
        self.__wmd_finder = GuiElementEmdFinder(self)

    @property
    def _finder(self):
        return self.__finder

    @property
    def _wmd_finder(self):
        return self.__wmd_finder

    @property
    def gns(self):
        '''
            Gui Namespace (GNS) object for this GuiElement.
        '''
        return self.__gns

    @property
    def root_element(self):
        '''
            Root Element for this GuiElement.
        '''
        return None

    def _find_element_with_js(self, js):
        raise Exception("With.JS is currently not supported for nested element finding.")

    def _find_multielement_with_js(self, js):
        raise Exception("With.JS is currently not supported for nested element finding.")

    def _wait_until_absent(self, wmd):
        try:
            self._wait_until_element_absent(wmd)
        except ArjunaTimeoutError:
            raise GuiWidgetPresentError(self.gui, wmd) 

    def wait_until_absent(self, *, fargs=None, **kwargs):
        '''
            Wait until a **GuiWidget** is absent inside this GuiElement.

            Keyword Arguments:
                fargs: A dictionary of key-value pairs for formatting the **GuiWidgetLocator**. Use **.format(**kwargs).wait_until_absent** for more Pythonic code when formatting.
                **kwargs: Arbitrary key-value pairs used to construct a **GuiWidgetLocator**

            Note:
                By default Wait is done until **ArjunaOption.GUIAUTO_MAX_WAIT** in the **Configuration** object associated with this **GuiElement**.

                You can pass **max_wait** argument to change this. Value is considered in seconds.
        '''
        from arjuna.tpi.guiauto.meta.locator import GuiWidgetLocator
        wmd = GuiWidgetLocator(fmt_args=fargs, **kwargs)._as_wmd()
        self._wait_until_absent(wmd)

    def contains(self, *, fargs=None, **kwargs):
        '''
            Check whether this GuiElement object contains a **GuiWidget**. Includes dynamic waiting.

            Keyword Arguments:
                fargs: A dictionary of key-value pairs for formatting the **GuiWidgetLocator**. Use **.format(**kwargs).wait_until_absent** for more Pythonic code when formatting.
                **kwargs: Arbitrary key-value pairs used to construct a **GuiWidgetLocator**

            Note:
                By default Wait is done until `ArjunaOption.GUIAUTO_MAX_WAIT` in the **Configuration** object associated with this **GuiElement**.

                You can pass **max_wait** argument to change this. Value is considered in seconds.
        '''
        try:
            self.element(fargs=fargs, **kwargs)
        except GuiWidgetNotPresentError:
            return False
        else:
            return True

    ########## Served by Template ########

    def formatter(self, **fargs) -> GuiWidgetLocatorFormatter:
        '''
            Create a :class:`~arjuna.tpi.guiauto.meta.formatter.GuiWidgetLocatorFormatter` object.

            Keyword Arguments:
                **fargs: Arbitrary key-value pairs to be used for formatting identifiers in **GuiWidgetLocator**.
        '''
        return GuiWidgetLocatorFormatter(self, **fargs)

    def locate(self, locator):
        '''
           Locate a `GuiWidget`.

           Arguments:
            locator: `GuiWidgetLocator` object.

            Returns:
                An object of type `GuiWidget`. Exact object type depends on the value of **type** attribute in **GuiWidgetLocator**. 
        '''
        return self._finder.locate(locator)

    def element(self, *, fargs=None, **kwargs):
        '''
            Locate a `GuiElement`.

            Keyword Arguments:
                fargs: A dictionary of key-value pairs for formatting the **GuiWidgetLocator**. Use **.format(kwargs).wait_until_absent** for more Pythonic code when formatting.
                **kwargs: Arbitrary key-value pairs used to construct a **GuiWidgetLocator**

            Returns:
                `GuiElement` object.
        '''
        return self._finder.element(fargs=fargs, **kwargs)

    def multi_element(self, fargs=None, **kwargs):
        '''
            Locate a `GuiMultiElement`.

            Keyword Arguments:
                fargs: A dictionary of key-value pairs for formatting the **GuiWidgetLocator**. Use **.format(kwargs).wait_until_absent** for more Pythonic code when formatting.
                **kwargs: Arbitrary key-value pairs used to construct a **GuiWidgetLocator**

            Returns:
                `GuiMultiElement` object.
        '''
        return self._finder.multi_element(fargs=fargs, **kwargs)

    def dropdown(self, fargs=None, **kwargs):
        '''
            Locate a `GuiDropDown`.

            Keyword Arguments:
                fargs: A dictionary of key-value pairs for formatting the **GuiWidgetLocator**. Use **.format(kwargs).wait_until_absent** for more Pythonic code when formatting.
                **kwargs: Arbitrary key-value pairs used to construct a **GuiWidgetLocator**

            Returns:
                `GuiDropDown` object.
        '''
        return self._finder.dropdown(fargs=fargs, **kwargs)

    def radio_group(self, fargs=None, **kwargs):
        '''
            Locate a `GuiRadioGroup`

            Keyword Arguments:
                fargs: A dictionary of key-value pairs for formatting the **GuiWidgetLocator**. Use **.format(kwargs).wait_until_absent** for more Pythonic code when formatting.
                **kwargs: Arbitrary key-value pairs used to construct a **GuiWidgetLocator**

            Returns:
                `GuiRadioGroup` object
        '''
        return self._finder.radio_group(fargs=fargs, **kwargs)


# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import threading
import types

from arjuna.core.reader.excel import *
from arjuna.core.reader.ini import *
from arjuna.core.reader.textfile import *
from arjuna.core.thread import decorators
from arjuna.tpi.error import *
from arjuna.core.error import *
from arjuna.core.types import constants
from arjuna.tpi.data.record import *
from arjuna.engine.data.record import *
# from arjuna.core.utils import sys_utils

class DataSource(metaclass=abc.ABCMeta):
    def __init__(self, *, context):
        super().__init__()
        self.lock = threading.RLock()
        self.name = None
        self.ended = False
        self.__context = context

    @property
    def context(self):
        return self.__context

    def get_name(self):
        return self.name

    def set_name(self, name):
        self.name = name

    @decorators.sync_method('lock')
    def terminate(self):
        self.ended = True

    def is_terminated(self):
        return self.ended

    def validate(self):
        pass

    @decorators.sync_method('lock')
    def next(self):
        if self.is_terminated():
            raise DataSourceFinished()

        try:
            data_record = self.get_next()
            if not data_record:
                raise StopIteration()
            record = self.process(data_record)
            if self.should_exclude(record):
                return self.next()
            else:
                return record
        except StopIteration:
            raise DataSourceFinished("Records Finished.")
        except Exception as e:
            import traceback
            traceback.print_exc()
            raise DataSourceFinished(
                "Problem happened in getting next record. No further records would be provided.")

    def process(self, data_record):
        return data_record

    @abc.abstractmethod
    def get_next(self):
        pass

    @abc.abstractmethod
    def should_exclude(self, data_record):
        return False

    @abc.abstractmethod
    def reset(self):
        pass

    @property
    def all_records(self):
        out = []
        while True:
            try:
                record = self.next()
                out.append(record)
            except DataSourceFinished as e:
                break
        return out


class FileDataSource(DataSource, metaclass=abc.ABCMeta):
    def __init__(self, path, *, context):
        super().__init__(context=context)
        self.__path = path
        self.__reader = None

    @property
    def path(self):
        return self.__path

    @property
    def reader(self):
        return self.__reader

    @reader.setter
    def reader(self, reader):
        self.__reader = reader

    def get_next(self):
        return self.reader.next()

    def should_exclude(self, data_record):
        return False

    def reset(self):
        self._load_file()

    @abc.abstractmethod
    def _load_file(self):
        pass

class DsvFileMapDataSource(FileDataSource):
    def __init__(self, path, delimiter="\t", *, context):
        super().__init__(path, context=context)
        self.__delimiter = delimiter
        self._load_file()

    def process(self, data_record):
        return DataRecord(context=self.context, **dict(data_record))

    def _load_file(self):
        self.reader = FileLine2MapReader(self.path, self.__delimiter)


class IniFileDataSource(FileDataSource):
    def __init__(self, path, *, context):
        super().__init__(path, context=context)
        if path.lower().endswith("ini"):
            self._load_file()
        else:
            raise Exception("Unsupported file extension for Ini reading.")

    def process(self, data_record):
        return DataRecord(context=self.context, **dict(data_record))

    def _load_file(self):
        self.reader = IniFile2MapReader(self.path)


class ExcelFileMapDataSource(FileDataSource):
    def __init__(self, path, *, context):
        super().__init__(path, context=context)
        if path.lower().endswith("xls"):
            self._load_file()
        else:
            raise Exception("Unsupported file extension for Excel reading.")

    def _load_file(self):
        self.reader = ExcelRow2MapReader(self.path)

    def should_exclude(self, data_record):
        return data_record._should_exclude()

    def process(self, data_record):
        return DataRecord(context=self.context, **dict(data_record))

class DummyDataSource(DataSource):

    def __init__(self, *, context):
        super().__init__(context=context)
        self.done = False

    def get_next(self):
        if self.done:
            raise DataSourceFinished()
        else:
            self.done = True
            return DataRecord(context=self.context)

    def should_exclude(self, data_record):
        return False


class SingleDataRecordSource(DataSource):
    def __init__(self, record, *, context):
        super().__init__(context=context)
        self.record = record
        self.done = False

    def get_next(self):
        if not self.is_terminated() and not self.done:
            self.done = True
            return self.record
        else:
            raise DataSourceFinished()

    def should_exclude(self, data_record):
        return False

    def reset(self):
        pass


class DataArrayDataSource(DataSource):
    def __init__(self, records, *, context):
        super().__init__(context=context)
        self.records = records
        self.__iter = iter(self.records)

    def get_next(self):
        return next(self.__iter)

    def should_exclude(self, data_record):
        return False

    def reset(self):
        self.__iter = iter(self.records)


class DataFunctionDataSource(DataSource):
    def __init__(self, func, *vargs, context, **kwargs):
        super().__init__(context=context)
        self.func = func
        self.vargs = vargs
        self.kwargs = kwargs
        try:
            self.reset()
        except:
            raise Exception("data_function should return an object that is an iterable.")

    def reset(self):
        self.__iter = iter(self.func(*self.vargs, **self.kwargs))

    def get_next(self):
        obj = next(self.__iter)
        if isinstance(obj, DataRecord):
            return obj
        elif type(obj) is tuple or type(obj) is list:
            return DataRecord(context=self.context, *obj)
        elif type(obj) is dict:
            return DataRecord(context=self.context, **obj)
        else:
            return DataRecord(obj, context=self.context)

    def terminate(self):
        super().terminate()
        # self.ds.terminate()

    def should_exclude(self, data_record):
        return False


class DataClassDataSource(DataSource):
    def __init__(self, dclass, *vargs, context, **kwargs):
        super().__init__(context=context)
        self.klass = dclass
        self.vargs = vargs
        self.kwargs = kwargs
        try:
            self.reset()
        except:
            raise Exception("data_class should implement Python iteration protocol.")

    def reset(self):
        self.__iter = iter(self.klass(*self.vargs, **self.kwargs))

    def get_next(self):
        obj = next(self.__iter)
        if isinstance(obj, DataRecord):
            return obj
        elif type(obj) is tuple or type(obj) is list:
            return DataRecord(context=self.context, *obj)
        elif type(obj) is dict:
            return DataRecord(context=self.context, **obj)
        else:
            return DataRecord(obj, context=self.context)

    def terminate(self):
        super().terminate()
        # self.ds.terminate()

    def should_exclude(self, data_record):
        return False


class MultiDataSource(DataSource):
    def __init__(self, dsource_defs, *, context):
        super().__init__(context=context)
        self.dsdefs = dsource_defs
        self.current_dsource = None
        self.def_iter = iter(self.dsdefs)

    def __init_next_source(self):
        try:
            dsdef = next(self.def_iter)
            self.current_dsource = dsdef.build(context=self.context)
        except:
            raise StopIteration()

    def reset(self):
        self.def_iter = iter(self.dsdefs)

    def get_next(self):
        if self.current_dsource is None:
            self.__init_next_source()

        try:
            return self.current_dsource.next()
        except StopIteration as e:
            self.__init_next_source()
            return self.get_next()

    def should_exclude(self, data_record):
        return False



# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import platform

import pytest

from arjuna.core.error import TestGroupsFinished
from arjuna.core.constant import *
from arjuna.tpi.constant import *


class TestGroup:

    def __init__(self, *, name, config, session, stage, rules=None):
        self.__name = name
        self.__session = session
        self.__stage = stage
        self.__config = config
        self.__thname = None
        self.__dry_run = session.dry_run
        self.__rules = rules
        self.__css_path = config.value(ArjunaOption.ARJUNA_ROOT_DIR) + "/arjuna/res/arjuna.css" 

    @classmethod
    def create_rule_strs(cls, include_exclude_dict):
        pickers_rulestr = {
            'ip': "package *= {}",
            'ep': "package *= {}",
            'im': "module *= {}",
            'em': "module *= {}",
            'it': "name *= {}",
            'et': "name *= {}",
        }

        rules = {'ir': [], 'er': []}

        for picker in pickers_rulestr:
            names = include_exclude_dict.pop(picker)
            if names:
                for name in names:
                    if picker.startswith('i'):
                        rules['ir'].append(pickers_rulestr[picker].format(name))
                    else:
                        rules['er'].append(pickers_rulestr[picker].format(name))
        return rules

    @property
    def config(self):
        return self.__config

    @property
    def thread_name(self):
        return self.__thname

    @thread_name.setter
    def thread_name(self, name):
        self.__thname = name

    @property
    def tests_dir(self):
        return self.__tests_dir

    def run(self):
        from arjuna import Arjuna
        from arjuna.tpi.constant import ArjunaOption
        Arjuna.register_group_params(name=self.__name, config=self.__config, thread_name=self.thread_name)
        self.__load_command_line()

        os.chdir(self.__project_dir)
        print("Executing pytest with args: {}".format(" ".join(self.__pytest_args)))


        pytest_retcode = pytest.main(self.__pytest_args)
        return pytest_retcode

    def __load_command_line(self):
        from arjuna import Arjuna
        from arjuna.tpi.constant import ArjunaOption
        self.__project_dir = self.config.value(ArjunaOption.PROJECT_ROOT_DIR)
        # import sys
        # sys.path.insert(0, self.__project_dir + "/..")
        self.__tests_dir = self.config.value(ArjunaOption.TESTS_DIR)
        suffix = ""
        if self.__name != "mgroup":
            suffix = "-" + self.__session.name + "-" + self.__stage.name + "-" + self.__name
        self.__xml_path = os.path.join(self.config.value(ArjunaOption.REPORT_XML_DIR), "report{}.xml".format(suffix))
        self.__html_path = os.path.join(self.config.value(ArjunaOption.REPORT_HTML_DIR), "report{}.html".format(suffix))
        self.__report_formats = self.config.value(ArjunaOption.REPORT_FORMATS)
        # self.__report_formats = Value.as_enum_list(rfmts, ReportFormat)
        res_path = os.path.abspath(os.path.join(os.path.dirname(os.path.realpath(__file__)), "../../res"))
        pytest_ini_path = res_path + "/pytest.ini"

        # -s is to print to console.
        self.__pytest_args = ["-c", pytest_ini_path, "--rootdir", self.__project_dir, "--disable-warnings", "-rxX", "--css", self.__css_path]
        if platform.system().casefold() == "Windows".casefold() :
            self.__pytest_args.extend(["--capture", "no"])
        else:
            self.__pytest_args.extend(["--no-print-logs", "--show-capture", "all"])
        self.__test_args = []
        self.__load_tests(rules=self.__rules)
        self.__load_meta_args()

    def __load_tests(self, *, rules):
        from arjuna import Arjuna
        from arjuna.engine.selection.selector import Selector
        selector = Selector()
        if rules:
            for rule in rules['ir']:
                selector.include(rule)
            for rule in rules['er']:
                selector.exclude(rule)
            

        Arjuna.register_test_selector_for_group(selector)

    def __load_meta_args(self):
        pytest_report_args = []

        if ReportFormat.XML in self.__report_formats:
            pytest_report_args.extend(["--junit-xml", self.__xml_path])

        if ReportFormat.HTML in self.__report_formats:
            pytest_report_args.extend(["--html", self.__html_path, "--self-contained-html"])

        self.__pytest_args.extend(pytest_report_args)
        self.__pytest_args.extend(self.__test_args)

        if self.__dry_run not in {False, None}:
            print("!!!!!! This is a DRY RUN !!!!!!!")
            if self.__dry_run == DryRunType.SHOW_TESTS:
                print("Dry Run Type: SHOW TESTS")
                print("You can see the test functions which will be executed as per settings of your command.")
                self.__pytest_args.extend(["--collect-only"])
            elif self.__dry_run == DryRunType.SHOW_PLAN:
                print("Dry Run Type: SHOW PLAN")
                print("You can see the test functions as well as the fixtures which will be executed as per settings of your command.")
                self.__pytest_args.extend(["--setup-plan"])
            elif self.__dry_run == DryRunType.CREATE_RES:
                print("Dry Run Type: CREATE RESOURCES")
                print("All resources will be created as per your current command. You can see the test functions which will be executed as per settings of your command.")
                self.__pytest_args.extend(["--setup-only"])

    def __str__(self):
        return "TestGroup: name:{}, config={}, rules={}".format(self.__name, self.config.name, self.__rules)


class YamlTestGroup(TestGroup):

    def __init__(self, *, name, group_yaml, session, stage):
        self.__config = stage.config
        self.__rules = {'ir': [], 'er': []}
        self.__process_yaml(group_yaml)
        super().__init__(name=name, config=self.__config, session=session, stage=stage, rules=self.__rules)

    def __process_yaml(self, group_yaml):
        from arjuna import Arjuna
        for gmd_name in group_yaml.section_names:
            pickers = {
                'ip': None,
                'ep': None,
                'im': None,
                'em': None,
                'it': None,
                'et': None,
            }
            gmd_name = gmd_name.lower()
            if gmd_name == "conf":
                self.__config = Arjuna.get_config(group_yaml["conf"])
            elif gmd_name in pickers:
                pickers[gmd_name] = group_yaml[gmd_name]
                if gmd_name.startswith('i'):
                    self.__rules['ir'].extend(TestGroup.create_rule_strs(pickers)['ir'])
                else:
                    self.__rules['er'].extend(TestGroup.create_rule_strs(pickers)['er'])
            elif gmd_name in {'ir', 'er'}:
                self.__rules[gmd_name].extend(group_yaml[gmd_name])


class MagicTestGroup(TestGroup):

    def __init__(self, *, session, stage, rules=None):
        super().__init__(name="mgroup", config=stage.config, session=session, stage=stage, rules=rules)


class TestGroups:

    def __init__(self):
        self.__names = []
        self.__iter = None

    def add_group(self, group):
        self.__names.append(group)

    def freeze(self):
        self.__iter = iter(self.__names)

    def __iter__(self):
        return self

    def next(self):
        try:
            return next(self.__iter)
        except StopIteration:
            raise TestGroupsFinished()

    def __str__(self):
        return str([str(c) for c in self.__names])


# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import platform

from browsermobproxy.exceptions import ProxyServerError

from .driverelement import SeleniumDriverElementDispatcher
from arjuna.interact.gui.dispatcher.driver.driver_commands import DriverCommands
from arjuna.interact.gui.dispatcher.driver.element_finder import ElementFinder
from arjuna.interact.gui.dispatcher.driver.melement import MultiElement
from selenium.webdriver.remote.webelement import WebElement

class SeleniumDriverDispatcher:

    def __init__(self):
        self.__config = None
        self.__driver = None
        self.__proxy_server = None
        self.__proxy = None
        self.__driver_service = None

    def __create_gui_element_dispatcher(self, element):
        return SeleniumDriverElementDispatcher.create_dispatcher(self, element)

    @property
    def driver(self):
        return self.__driver

    @property
    def proxy(self):
        return self.__proxy

    def launch(self, config):
        self.__config = config
        from .browser_launcher import BrowserLauncher

        svc_url = config["arjuna_options"]["SELENIUM_SERVICE_URL"]
        driver_download = config["arjuna_options"]["SELENIUM_DRIVER_DOWNLOAD"]
        browser_name = config["arjuna_options"]["BROWSER_NAME"]
        driver_path = config["arjuna_options"]["SELENIUM_DRIVER_PATH"]
        if svc_url.lower() == "not_set":
            from arjuna.tpi.constant import BrowserName
            driver_service = None
            driver_downloader = None
            if browser_name == BrowserName.CHROME:
                from selenium.webdriver.chrome.service import Service
                driver_service = Service
                from webdriver_manager.chrome import ChromeDriverManager
                driver_downloader = ChromeDriverManager
            elif browser_name == BrowserName.FIREFOX:
                from selenium.webdriver.firefox.service import Service
                driver_service = Service
                from webdriver_manager.firefox import GeckoDriverManager
                driver_downloader = GeckoDriverManager
            if driver_download:
                driver_path = driver_downloader().install()
            self.__driver_service = Service(driver_path)
            self.__driver_service.start()
            svc_url = self.__driver_service.service_url
        else:
            if not svc_url.lower().endswith("/wd/hub"):
                svc_url += "/wd/hub"

        # BrowserMob
        from arjuna import Arjuna
        bmproxy_server = Arjuna._get_bmproxy_server()
        if bmproxy_server is not None:
            self.__proxy = bmproxy_server.create_proxy()

        self.__driver = BrowserLauncher.launch(config, svc_url=svc_url, proxy=self.__proxy) 

    def quit(self):
        DriverCommands.quit(self.__driver)
        if self.__driver_service:
            self.__driver_service.stop()
        if self.__proxy:
            self.__proxy.close()

    def go_to_url(self, url):
        DriverCommands.go_to_url(self.__driver, url)

    def go_back_in_browser(self):
        DriverCommands.go_back_in_browser(self.__driver)

    def go_forward_in_browser(self):
        DriverCommands.go_forward_in_browser(self.__driver)

    def refresh_browser(self):
        DriverCommands.refresh_browser(self.__driver)

    def get_source(self):
        return DriverCommands.get_source(self.__driver)

    def send_keys(self, key_str):
        DriverCommands.send_keys(self.__driver, key_str)

    def execute_javascript(self, script, *args):
        return DriverCommands.execute_javascript(self.__driver, script, 
                *[
                isinstance(arg, SeleniumDriverElementDispatcher) and arg.driver_element or arg for arg in args
            ]
        )

    def take_screenshot(self, file_path):
        DriverCommands.take_screenshot(self.__driver, file_path)

    def take_screenshot_as_base64(self):
        return DriverCommands.take_screenshot_as_base64(self.__driver)

    def find_element(self, with_type, with_value):
        element = ElementFinder.find_element(self.__driver, with_type, with_value)
        return 1, self.__create_gui_element_dispatcher(element)

    def __process_single_js_element(self, element):
        # JS returns null, undefined
        if element is None:
            raise Exception("JavaScript could not find element.")
        elif not isinstance(element, WebElement):
            raise Exception("JavaScript returned a non-element object.")
        else:
            return element

    def __process_js_element_list(self, elements):
        if not elements: raise Exception("JavaScript could not find element.")
        return [self.__process_single_js_element(e) for e in elements]

    def __process_js_element(self, element):
        if type(element) is list:
            element = self.__process_js_element_list(element)[0]
        else:
            element = self.__process_single_js_element(element)
        return element

    def __process_js_multielement(self, elements):
        if type(elements) is list:
            elements = self.__process_js_element_list(elements)
        else:
            elements = [self.__process_single_js_element(elements)]
        return elements

    def find_element_with_js(self, js):
        element = self.execute_javascript(js)
        element = self.__process_js_element(element)
        return 1, self.__create_gui_element_dispatcher(element)

    def find_multielement(self, with_type, with_value):
        web_elements = ElementFinder.find_elements(self.__driver, with_type, with_value)
        melement = MultiElement([SeleniumDriverElementDispatcher(self, web_element) for web_element in web_elements])
        return melement.get_size(), melement

    def find_multielement_with_js(self, js):
        web_elements = self.execute_javascript(js)
        web_elements = self.__process_js_multielement(web_elements)
        melement = MultiElement([SeleniumDriverElementDispatcher(self, web_element) for web_element in web_elements])
        return melement.get_size(), melement

    def get_current_window_handle(self):
        return DriverCommands.get_current_window_handle(self.__driver)

    def get_current_window_title(self):
        return DriverCommands.get_window_title(self.__driver)

    def maximize_current_window(self):
        DriverCommands.maximize_window(self.__driver)

    def get_current_window_size(self):
        res = DriverCommands.get_current_window_size(self.__driver)
        return {"width" : res[0], "height" : res[1]}

    def get_all_window_handles(self):
        return DriverCommands.get_all_winodw_handles(self.__driver)

    def focus_on_window(self, handle):
        DriverCommands.focus_on_window(self.__driver, handle)

    def close_current_window(self):
        DriverCommands.close_current_window(self.__driver)

    def is_web_alert_present(self):
        return DriverCommands.is_web_alert_present(self.__driver)

    def confirm_web_alert(self):
        DriverCommands.confirm_web_alert(self.__driver)

    def dismiss_web_alert(self):
        DriverCommands.dismiss_web_alert(self.__driver)

    def get_text_from_web_alert(self):
        return DriverCommands.get_text_from_web_alert(self.__driver)

    def send_text_to_web_alert(self, text):
        DriverCommands.send_text_to_web_alert(self.__driver, text)

    def focus_on_frame(self, element_dispatcher):
        DriverCommands.focus_on_frame(self.__driver, element_dispatcher.driver_element)

    def get_element_for_setu_id(self, id):
        return self.__driver_elements[id]

    def focus_on_parent_frame(self):
        DriverCommands.focus_on_parent_frame(self.__driver)

    def focus_on_dom_root(self):
        DriverCommands.focus_on_dom_root(self.__driver)

    def perform_action_chain(self, action_chain):
        DriverCommands.perform_action_chain(self, self.__driver, action_chain)

    def hover_on_element(self, element_dispatcher):
        DriverCommands.hover_on_element(self.__driver, element_dispatcher.driver_element)

    def mouse_click_on_element(self, element_dispatcher):
        DriverCommands.mouse_click_on_element(self.__driver, element_dispatcher.driver_element)

    def scroll_to_element(self, element_dispatcher):
        DriverCommands.scroll_to_element(self.__driver, element_dispatcher.driver_element)





# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from arjuna.core.utils import data_utils
from arjuna.core.types import constants


class FileReader:
    def __init__(self, file_path):
        self._fpath = file_path
        self._f = open(self._fpath, "r")

    def close(self):
        self._f.close()

    def read(self):
        return self._f.read()

    def _get_file(self):
        return self._f


class TextResourceReader(FileReader):
    def __init__(self, file_name):
        super().__init__(
        os.path.abspath(os.path.join(os.path.dirname(__file__), os.sep.join(["../..", "res", file_name]))))


class FileLineReader:
    def __init__(self, file_path):
        self.__filereader = open(file_path, "r")

    def __iter__(self):
        return self

    def next(self):
        l = self.__filereader.readline()
        if not l:
            self.close()
            raise StopIteration()
        elif l.strip().startswith('#'):
            return self.next()
        else:
            return l.rstrip()

    def close(self):
        self.__filereader.close()


class FileLine2ArrayReader:
    def __init__(self, file_path, delimiter="\t"):
        self.__filereader = FileLineReader(file_path)
        self.delimiter = delimiter
        self.headers = None
        self._populate_headers()

    def _populate_headers(self):
        try:
            self.headers = self.next()
        except:
            raise Exception("Invalid input file data. Empty headers line.")

    def next(self):
        line = self.__filereader.next()
        return data_utils.split(line.rstrip(), self.delimiter)

    def get_headers(self):
        return self.headers

    def close(self):
        self.__filereader.close()


class FileLine2MapReader:
    def __init__(self, file_path, delimiter="\t"):
        self.__filereader = FileLine2ArrayReader(file_path, delimiter)

    def get_headers(self):
        return self.__filereader.headers

    def close(self):
        self.__filereader.close()

    def next(self):
        line_parts = self.__filereader.next()

        if len(self.get_headers()) != len(line_parts):
            raise Exception(
                "Invalid input file data. Number of headers and data values do not match.{0}Headers:{1}{0}Data values:{2}{0}".format(
                    os.linesep, self.get_headers(), line_parts))
        else:
            return dict(zip(self.get_headers(), line_parts))



# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import abc
import random

from arjuna.tpi.helper.arjtype import CIStringDict
from .record import DataRecord

def _get_file_name(path):
    name = os.path.basename(path)
    return os.path.splitext(name)[0]

class IndexedDataReference(metaclass=abc.ABCMeta):
    '''
        Base class for autoloaded Indexed Data References in Arjuna.

        Arguments:
            pydict: A python repr of a list with indices as keys.

        Note:
            It behaves like a Python tuple. So, you get items by index and loop over it.
    '''

    def __init__(self, pydict):
        self.__records = pydict
        self.__iter = None

    def __len__(self):
        return len(self.__records)

    def __getitem__(self, index):
        return self._record_for(index)

    def first(self):
        '''
        First item in this data reference.
        '''
        return self[0]

    def random(self):
        '''
        Random item in this data reference.
        '''
        index = random.randint(0, len(self) -1)
        return self[index]

    def last(self):
        '''
        Last item in this data reference.
        '''
        return self[-1]

    def _record_for(self, index):
        try:
            # To support negative indices
            indices = list(self.__records.keys())
            try:
                target_index = indices[index]
            except:
                raise KeyError()
            else:
                return self.__records[target_index]
        except KeyError:
            raise Exception("Index {} not found in data reference: {}.".format(index, self.__class__.__name__))

    def __str__(self):
        return str({k: str(v) for k,v in self.__records.items()})

    def enumerate(self):
        '''
        Print all items in this data reference.
        '''
        for k,v in self.__records.items():
            print(k, "::", type(v), str(v))

    def __iter__(self):
        self.__iter = iter(self.__records.values())
        return self

    def __next__(self):
        return next(self.__iter)


class ContextualDataReference(metaclass=abc.ABCMeta):
    '''
        Base class for autoloaded Contextual data references in Arjuna.

        Arguments:
            path: Path of the contextual data reference file.

        Note:
            It behaves like a Python dictionary. So, you get items by names/keys and loop over it.
    '''

    def __init__(self, path):
        self.__path = path
        self.__name = _get_file_name(path)
        self.__map = CIStringDict()
        self._populate()
        self.__iter = None

    def __iter__(self):
        self.__iter = iter(self.__map)
        return self

    def __next__(self):
        return next(self.__iter)

    def __len__(self):
        return len(self.__map)

    def __getitem__(self, context):
        return self._record_for(context)

    def keys(self):
        '''
        Names of contexts/keys in this reference.
        '''
        return self.__map.keys()

    def items(self):
        '''
        Items iterator for this reference.
        '''
        return self.__map.items()

    @property
    def _map(self):
        return self.__map

    @property
    def path(self):
        '''
        Path of this reference file.
        '''
        return self.__path

    @property
    def name(self):
        '''
        Name of this reference.
        '''
        return self.__name

    def _update(self, data_reference):
        for context, record in data_reference.map.items():
            if context not in self.map:
                self.__map[context] = CIStringDict()
            self.__map[context].update(record.named_values)

    def _update_from_dict(self, context, map):
        if context not in self.map:
            self.__map[context] = dict()
        self.__map[context].update(map)

    def _add_record(self, context, record):
        self.__map[context] = DataRecord(context="Ref-{}[{}]".format(self.name, context), **record)

    def _record_for(self, context):
        try:
            return self.__map[context]
        except:
            raise Exception("{} at {} does not contain {} context key.".format(self.__class__.__name__, self.path, context))

    def __str__(self):
        return str({k: str(v) for k,v in self.__map.items()})

    def enumerate(self):
        '''
        Print all items in this data reference.
        '''
        for k,v in self.__map.items():
            print(k, "::", type(v), str(v))

    @abc.abstractmethod
    def _populate(self):
        pass


# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
Data Driven Testing Markup

Provides data source builder classes that can be provided to drive_with argument of @test decorator.

The names of the classes have been kept in lower case for aesthetic purpose of the @test decorator.
'''

from arjuna.engine.data.source import *
from arjuna.engine.data.factory import *
from arjuna.tpi.constant import *
from arjuna.core.constant import *
from arjuna.core.utils import obj_utils
from arjuna.engine.data.factory import *

class _DataMarkUp:
    pass

class record(_DataMarkUp):
    '''
        Data Record Data Source

        Args:
            vargs: Any number of objects
            *kwargs: Arbitrary keywrod arguments
    '''

    from arjuna.engine.data.source import SingleDataRecordSource
    def __init__(self, *vargs, **kwargs):
        super().__init__()
        self.__vargs = vargs
        self.__kwargs = kwargs
        if not vargs and not kwargs:
            raise Exception("No data provided in Data") 

    def build(self, context="Test") -> 'DataSource':
        '''
            Create corresponding DataSource
        '''
        return SingleDataRecordSource(self.get_record(context=context), context=context)

    def get_record(self, *, context="Test"):
        return DataRecord(context=context, *self.__vargs, **self.__kwargs)

class records(_DataMarkUp):
    '''
        Multiple Data Record Data Source

        Args:
            *records: Any number of record objects
    '''

    from arjuna.engine.data.source import DataArrayDataSource
    def __init__(self, *records):
        super().__init__()
        for i in records:
            if not isinstance(i, record):
                raise Exception("Items in records() must be instances of record.")
        self.__records = records

    def build(self, context="Test") -> 'DataSource':
        '''
            Create corresponding DataSource
        '''
        return DataArrayDataSource(self.get_records(context=context), context=context)

    def get_records(self, *, context="Test"):
        return [i.get_record(context=context) for i in self.__records]

class data_function(_DataMarkUp):
    '''
        Data Function/Generator Data Source

        Args:
            func: Function object to be called.
            vargs: Any number of objects to be passed to function when calling it.
            *kwargs: Arbitrary keywrod arguments to be passed to function when calling it.
    '''

    def __init__(self, func, *vargs, **kwargs):
        super().__init__()
        if not obj_utils.callable(func):
            raise Exception("Argument provided to data_function should be a function object.")
        elif not obj_utils.is_function(func):
            raise Exception("Argument for data_function should be a function object which does not exist inside a class.")
        self.func = func
        self.vargs = vargs
        self.kwargs = kwargs

    def build(self, context="Test") -> 'DataSource':
        '''
            Create corresponding DataSource
        '''
        return DataFunctionDataSource(self.func, *self.vargs, context=context, **self.kwargs)


class data_class(_DataMarkUp):
    '''
        Data Class Data Source

        Args:
            dsclass: Class to be instantiated
            vargs: Any number of objects for instantiating the class.
            *kwargs: Arbitrary keywrod arguments  for instantiating the class.
    '''

    def __init__(self, dsclass, *vargs, **kwargs):
        super().__init__()
        if not isinstance(type(dsclass), type):
            raise Exception("Data Class expects dsclass to be a class.")
        self.dsclass = dsclass
        self.vargs = vargs
        self.kwargs = kwargs

    def build(self, context="Test") -> 'DataSource':
        '''
            Create corresponding DataSource
        '''
        return DataClassDataSource(self.dsclass, *self.vargs, context=context, **self.kwargs)


class data_file(_DataMarkUp):
    '''
        Data File Data Source

        Args:
            path: Path of the file
            
        Keyword Arguments:
            delimiter: (Optional) Delimiter to be used for text files. Default is tab (\t)
    '''
    def __init__(self, path=None, *, delimiter="\t"):
        super().__init__()
        self.path = path
        self.delimiter = delimiter

    def build(self, context="Test") -> 'DataSource':
        '''
            Create corresponding DataSource
        '''
        source = create_file_data_source(self.path, delimiter=self.delimiter, context=context)
        return source


class many_data_sources(_DataMarkUp):
    '''
        Multiple Data Source

        Args:
            *dsources: Any Arjuna Data Source
            
        Keyword Arguments:
            delimiter: (Optional) Delimiter to be used for text files. Default is tab (\t)
    '''
    def __init__(self, *dsources):
        super().__init__()
        if not dsources:
            raise Exception("You must provide atleast one data source as argument for many_data_sources")
        for dsource in dsources:
            if not isinstance(dsource, _DataMarkUp):
                raise Exception("All arguments of many_data_sources must be data sources.")
        self.dsource_defs = dsources

    def build(self, context="Test") -> 'DataSource':
        '''
            Create corresponding DataSource
        '''
        return MultiDataSource(self.dsource_defs, context=context)






# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from setuptools import setup, find_packages

packages = find_packages()

this_directory = os.path.abspath(os.path.dirname(os.path.realpath(__file__)))
with open(os.path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name = "arjuna",
    version = "1.1.7",
    url = "https://rahulverma.net",
    description = "Arjuna is a Python based test automation framework developed by Rahul Verma (www.rahulverma.net).",
    author = "Rahul Verma",
    author_email = "",
    packages = packages,
    long_description=long_description,
    long_description_content_type='text/markdown',
    package_data = {
        '' :  [
                    "*.txt",
                    "*.md",
                    "*.cfg",
        ],
        'arjuna' : [
                    "res/*.xml",
                    "res/*.help",
                    "res/*.txt",
                    "res/*.conf",
                    "res/*.ini",
                    "res/*.py",
                    "res/*.yaml",
                    "res/*.css",
                    "res/*.js",
                    "res/*.html"
                ]
    },
    install_requires = ["lxml", "requests", "requests-toolbelt", "selenium", "webdriver_manager", "xlrd", "xlwt", "pyparsing", "pyhocon", "pytest", "pytest-html", "pytest-dependency", "PyYAML", "mimesis", "jsonpath-rw", "jsonpath-rw-ext", "genson", "jsonschema", "Pallets-Sphinx-Themes", "oauthlib", "requests_oauthlib", "bs4", "browsermob-proxy", "haralyzer"],
    keywords = "arjuna selenium testing automation page-object data-driven",
    license = "Apache License, Version 2.0",
    classifiers=[
    'Environment :: Console',
    'Development Status :: 4 - Beta',
    'Intended Audience :: Developers',
    'Intended Audience :: Information Technology',
    'Topic :: Software Development :: Quality Assurance',
    'Topic :: Software Development :: Testing',
    'License :: OSI Approved :: Apache Software License',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3 :: Only',
    'Programming Language :: Python :: Implementation :: CPython',
    'Operating System :: OS Independent',
    'Natural Language :: English'
    ]
)



# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from selenium.webdriver.common.action_chains import ActionChains

class DriverCommands:

    @classmethod
    def go_to_url(cls, driver, url):
        driver.get(url)

    @classmethod
    def refresh_browser(cls, driver):
        driver.refresh()

    @classmethod
    def go_back_in_browser(cls, driver):
        driver.back()

    @classmethod
    def go_forward_in_browser(cls, driver):
        driver.forward()

    @classmethod
    def quit(cls, driver):
        driver.quit()

    @classmethod
    def get_page_title(cls, driver):
        return driver.title

    @classmethod
    def get_url(cls, driver):
        return driver

    @classmethod
    def get_source(cls, driver):
        return driver.page_source

    @classmethod
    def send_keys(cls, driver, key_str):
        print(key_str)
        ActionChains(driver).send_keys(key_str).perform()

    @classmethod
    def is_web_alert_present(cls, driver):
        from selenium.webdriver.support.ui import WebDriverWait
        from selenium.webdriver.support import expected_conditions as EC
        try:
            WebDriverWait(driver, 1).until(EC.alert_is_present(),'Timed out.')
            return True
        except Exception as e:
            return False

    @classmethod
    def confirm_web_alert(cls, driver):
        driver.switch_to.alert.accept()

    @classmethod
    def dismiss_web_alert(cls, driver):
        driver.switch_to.alert.dismiss()

    @classmethod
    def send_text_to_web_alert(cls, driver, text):
        driver.switch_to.alert.send_keys(text)

    @classmethod
    def get_text_from_web_alert(cls, driver):
        return driver.switch_to.alert.text

    @classmethod
    def focus_on_frame(cls, driver, element):
        driver.switch_to.frame(element)

    @classmethod
    def focus_on_dom_root(cls, driver):
        return driver.switch_to.default_content()

    @classmethod
    def focus_on_parent_frame(cls, driver):
        driver.switch_to.parent_frame()

    @classmethod
    def execute_javascript(cls, driver, script, *args):
        from arjuna import log_debug
        log_debug("Executing JavaScript {} with args {}.".format(script, args))
        return driver.execute_script(script, *args)

    @classmethod
    def take_screenshot(cls, driver, file_path):
        return driver.save_screenshot(file_path)

    @classmethod
    def take_screenshot_as_base64(cls, driver):
        return driver.get_screenshot_as_base64()

    @classmethod
    def set_window_size(cls, driver, width, height):
        driver.set_window_size(width, height)

    @classmethod
    def maximize_window(cls, driver):
        driver.maximize_window()

    @classmethod
    def get_current_window_handle(cls, driver):
        return driver.current_window_handle

    @classmethod
    def focus_on_window(cls, driver, window_handle):
        driver.switch_to.window(window_handle)

    @classmethod
    def close_current_window(cls, driver):
        driver.close()

    @classmethod
    def get_window_title(cls, driver):
        return driver.title

    @classmethod
    def get_current_window_size(cls, driver):
        return driver.get_window_size()

    @classmethod
    def get_all_winodw_handles(cls, driver):
        return driver.window_handles

    @classmethod
    def replace_with_element(cls, setu_driver, value_tuple):
        if value_tuple[1] == True:
            return setu_driver.get_element_for_setu_id(value_tuple[0])
        else:
            return value_tuple[0]
        
    @classmethod
    def perform_action_chain(cls, setu_driver, driver, action_chain):
        chain = ActionChains(driver)
        for action in action_chain:
            kwargs = {k:cls.replace_with_element(setu_driver, v) for k,v in action[1].items()}
            getattr(chain, action[0])(**kwargs)
        chain.perform()

    @classmethod
    def hover_on_element(cls, driver, webelement):
        chain = ActionChains(driver).move_to_element(webelement).perform()

    @classmethod
    def mouse_click_on_element(cls, driver, webelement):
        chain = ActionChains(driver).click(webelement).perform()

    @classmethod
    def scroll_to_element(cls, driver, webelement):
        cls.execute_javascript(driver, "arguments[0].scrollIntoView(true);", webelement)



# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

class DriverElementCommands:

    @classmethod
    def send_text(cls, element, text):
        element.send_keys(text)

    @classmethod
    def clear_text(cls, element):
        element.clear()

    @classmethod
    def submit(cls, element):
        element.submit()

    @classmethod
    def click(cls, element):
        element.click()

    @classmethod
    def get_text_content(cls, element):
        return element.text

    @classmethod
    def get_attr_value(cls, element, attr):
        return element.get_attribute(attr)

    @classmethod
    def get_tag_name(cls, element):
        return element.tag_name

    @classmethod
    def is_visible(cls, element):
        return element.is_displayed()

    @classmethod
    def is_clickable(cls, element):
        return element.is_enabled()

    @classmethod
    def is_selected(cls, element):
        return element.is_selected()